From 058782c3ed79ff92c2c5a0c484e1b6957134a437 Mon Sep 17 00:00:00 2001
From: Tomas Vondra <tomas@2ndquadrant.com>
Date: Sun, 10 Dec 2017 20:06:06 +0100
Subject: [PATCH 3/3] disable unbounded hashtable growth

---
 src/include/lib/simplehash.h | 14 ++++++++++++--
 1 file changed, 12 insertions(+), 2 deletions(-)

diff --git a/src/include/lib/simplehash.h b/src/include/lib/simplehash.h
index 4f687a8..20c8d98 100644
--- a/src/include/lib/simplehash.h
+++ b/src/include/lib/simplehash.h
@@ -171,6 +171,8 @@ SH_SCOPE void SH_STAT(SH_TYPE * tb);
 #endif
 /* increase fillfactor if we otherwise would error out */
 #define SH_MAX_FILLFACTOR (0.98)
+/* don't grow the hash table too aggressively */
+#define SH_MIN_FILLFACTOR (0.33)
 /* grow if actual and optimal location bigger than */
 #ifndef SH_GROW_MAX_DIB
 #define SH_GROW_MAX_DIB 25
@@ -501,6 +503,7 @@ SH_INSERT(SH_TYPE * tb, SH_KEY_TYPE key, bool *found)
 	uint32		curelem;
 	SH_ELEMENT_TYPE *data;
 	uint32		insertdist;
+	bool		allow_grow;
 
 restart:
 	insertdist = 0;
@@ -528,6 +531,13 @@ restart:
 		/* SH_STAT(tb); */
 	}
 
+	/*
+	 * We don't want the fill factor to drop below SH_MIN_FILLFACTOR, and
+	 * growing the table doubles the size. So we only allow it if the
+	 * current fill factor is twice SH_MIN_FILLFACTOR.
+	 */
+	allow_grow = (tb->members >= 2 * SH_MIN_FILLFACTOR * tb->size);
+
 	/* perform insert, start bucket search at optimal location */
 	data = tb->data;
 	startelem = SH_INITIAL_BUCKET(tb, hash);
@@ -599,7 +609,7 @@ restart:
 				 * imbalance is filling a (currently) small table, from a
 				 * currently big one, in hash-table order.
 				 */
-				if (++emptydist > SH_GROW_MAX_MOVE)
+				if (allow_grow && (++emptydist > SH_GROW_MAX_MOVE))
 				{
 					tb->grow_threshold = 0;
 					goto restart;
@@ -646,7 +656,7 @@ restart:
 		 * likely cause of such imbalance is filling a (currently) small
 		 * table, from a currently big one, in hash-table order.
 		 */
-		if (insertdist > SH_GROW_MAX_DIB)
+		if (allow_grow && (insertdist > SH_GROW_MAX_DIB))
 		{
 			tb->grow_threshold = 0;
 			goto restart;
-- 
2.9.5

